__all__ = ['__version__', 'get_version']

version_info = (1, 5, 0, 0, 3)
# format: ('mujoco_major', 'mujoco_minor', 'mujoco_py_major',
# 'mujoco_py_major', 'mujoco_py_patch')


def get_version():
    "Returns the version as a human-format string."
    return '%d.%d.%d.%d.%d' % version_info

__version__ = get_version()



import copy
from os.path import join, expanduser

import numpy as np



def remove_empty_lines(string):
    lines = []
    for line in string.splitlines():
        if line.strip():
            lines.append(line)
    return "\n".join(lines)


def rec_assign(node, assign):
    # Assigns values to node recursively.
    # This is neccessary to avoid overriding pointers in MuJoCo.
    for field in dir(node):
        if field.find("__") == -1:
            val = getattr(node, field)
            if isinstance(val, (int, bool, float, None.__class__, str)):
                setattr(node, field, assign[field])
            elif isinstance(val, np.ndarray):
                val[:] = assign[field][:]
            elif not hasattr(val, "__call__"):
                rec_assign(val, assign[field])


def rec_copy(node):
    # Recursively copies object to dictionary.
    # Applying directly copy.deepcopy causes seg fault.
    ret = {}
    for field in dir(node):
        if field.find("__") == -1:
            val = getattr(node, field)
            if isinstance(val, (int, bool, float, None.__class__, str)):
                ret[field] = val
            elif isinstance(val, np.ndarray):
                ret[field] = copy.deepcopy(val)
            elif not hasattr(val, "__call__"):
                ret[field] = rec_copy(val)
    return ret


def discover_mujoco():
    """
    Discovers where MuJoCo is located in the file system.
    Currently assumes path is in ~/.mujoco

    Returns:
    - mjpro_path (str): Path to MuJoCo Pro 1.50 directory.
    - key_path (str): Path to the MuJoCo license key.
    """
    key_path = join(expanduser('~'), '.mujoco', 'mjkey.txt')
    mjpro_path = join(expanduser('~'), '.mujoco', 'mjpro150')
    return (mjpro_path, key_path)



from threading import Lock
from pymj.generated import const

cdef class MjRenderContext(object):
    """
    Class that encapsulates rendering functionality for a
    MuJoCo simulation.
    """

    cdef mjModel *_model_ptr
    cdef mjData *_data_ptr

    cdef mjvScene _scn
    cdef mjvCamera _cam
    cdef mjvOption _vopt
    cdef mjvPerturb _pert
    cdef mjrContext _con

    # Public wrappers
    cdef readonly PyMjvScene scn
    cdef readonly PyMjvCamera cam
    cdef readonly PyMjvOption vopt
    cdef readonly PyMjvPerturb pert
    cdef readonly PyMjrContext con

    cdef readonly object _opengl_context
    cdef readonly int _visible
    cdef readonly list _markers
    cdef readonly dict _overlay

    cdef readonly bint offscreen
    cdef public object sim

    def __cinit__(self):
        maxgeom = 1000
        mjv_makeScene(&self._scn, maxgeom)
        mjv_defaultCamera(&self._cam)
        mjv_defaultOption(&self._vopt)
        mjr_defaultContext(&self._con)

    def __init__(self, MjSim sim, bint offscreen=True):
        self.sim = sim
        self._setup_opengl_context(offscreen)
        self.offscreen = offscreen

        # Ensure the model data has been updated so that there
        # is something to render
        sim.forward()

        sim.add_render_context(self)

        self._model_ptr = sim.model.ptr
        self._data_ptr = sim.data.ptr
        self.scn = WrapMjvScene(&self._scn)
        self.cam = WrapMjvCamera(&self._cam)
        self.vopt = WrapMjvOption(&self._vopt)
        self._pert.active = 0
        self._pert.select = 0
        self.pert = WrapMjvPerturb(&self._pert)

        self._markers = []
        self._overlay = {}

        self._init_camera(sim)
        self._set_mujoco_buffers()

    def update_sim(self, MjSim new_sim):
        self._model_ptr = new_sim.model.ptr
        self._data_ptr = new_sim.data.ptr
        self._set_mujoco_buffers()
        for render_context in self.sim.render_contexts:
            new_sim.add_render_context(render_context)
        self.sim = new_sim

    def _set_mujoco_buffers(self):
        mjr_makeContext(self._model_ptr, &self._con, mjFONTSCALE_150)
        if self.offscreen:
            mjr_setBuffer(mjFB_OFFSCREEN, &self._con);
            if self._con.currentBuffer != mjFB_OFFSCREEN:
                raise RuntimeError('Offscreen rendering not supported')
        else:
            mjr_setBuffer(mjFB_WINDOW, &self._con);
            if self._con.currentBuffer != mjFB_WINDOW:
                raise RuntimeError('Window rendering not supported')
        self.con = WrapMjrContext(&self._con)

    def _setup_opengl_context(self, offscreen):
        if not offscreen or sys.platform == 'darwin':
            self._opengl_context = GlfwContext(offscreen=offscreen)
        else:
            self._opengl_context = OffscreenOpenGLContext()

    def _init_camera(self, sim):
        # Make the free camera look at the scene
        self.cam.type = const.CAMERA_FREE
        self.cam.fixedcamid = -1
        for i in range(3):
            self.cam.lookat[i] = sim.model.stat.center[i]
        self.cam.distance = sim.model.stat.extent

    def render(self, width, height, camera_id=None):
        cdef mjrRect rect
        rect.left = 0
        rect.bottom = 0
        rect.width = width
        rect.height = height

        # Sometimes buffers are too small.
        if width > self._con.offWidth or height > self._con.offHeight:
            self._model_ptr.vis.global_.offwidth = max(width, self._model_ptr.vis.global_.offwidth)
            self._model_ptr.vis.global_.offheight = max(height, self._model_ptr.vis.global_.offheight)
            mjr_freeContext(&self._con)
            self._set_mujoco_buffers()

        if camera_id is not None:
            if camera_id == -1:
                self.cam.type = const.CAMERA_FREE
            else:
                self.cam.type = const.CAMERA_FIXED
            self.cam.fixedcamid = camera_id

        self._opengl_context.set_buffer_size(width, height)

        mjv_updateScene(self._model_ptr, self._data_ptr, &self._vopt,
                        &self._pert, &self._cam, mjCAT_ALL, &self._scn)

        for marker_params in self._markers:
            self._add_marker_to_scene(marker_params)

        mjr_render(rect, &self._scn, &self._con)
        for gridpos, (text1, text2) in self._overlay.items():
            mjr_overlay(const.FONTSCALE_150, gridpos, rect, text1.encode(), text2.encode(), &self._con)

    def read_pixels(self, width, height, depth=True):
        cdef mjrRect rect
        rect.left = 0
        rect.bottom = 0
        rect.width = width
        rect.height = height

        rgb_arr = np.zeros(3 * rect.width * rect.height, dtype=np.uint8)
        depth_arr = np.zeros(rect.width * rect.height, dtype=np.float32)
        cdef unsigned char[::view.contiguous] rgb_view = rgb_arr
        cdef float[::view.contiguous] depth_view = depth_arr
        mjr_readPixels(&rgb_view[0], &depth_view[0], rect, &self._con)
        rgb_img = np.flipud(rgb_arr.reshape(rect.height, rect.width, 3))
        if depth:
            depth_img = np.flipud(depth_arr.reshape(rect.height, rect.width))
            return (rgb_img, depth_img)
        else:
            return rgb_img

    def upload_texture(self, int tex_id):
        """ Uploads given texture to the GPU. """
        self._opengl_context.make_context_current()
        mjr_uploadTexture(self._model_ptr, &self._con, tex_id)

    def draw_pixels(self, np.ndarray[np.uint8_t, ndim=3] image, int left, int bottom):
        """Draw an image into the OpenGL buffer."""
        cdef unsigned char[::view.contiguous] image_view = image.ravel()
        cdef mjrRect viewport
        viewport.left = left
        viewport.bottom = bottom
        viewport.width = image.shape[1]
        viewport.height = image.shape[0]
        mjr_drawPixels(&image_view[0], NULL, viewport, &self._con)

    def move_camera(self, int action, double reldx, double reldy):
        """ Moves the camera based on mouse movements. Action is one of mjMOUSE_*. """
        mjv_moveCamera(self._model_ptr, action, reldx, reldy, &self._scn, &self._cam)

    def add_overlay(self, int gridpos, str text1, str text2):
        """ Overlays text on the scene. """
        if gridpos not in self._overlay:
            self._overlay[gridpos] = ["", ""]
        self._overlay[gridpos][0] += text1 + "\n"
        self._overlay[gridpos][1] += text2 + "\n"

    def add_marker(self, **marker_params):
        self._markers.append(marker_params)

    def _add_marker_to_scene(self, marker_params):
        """ Adds marker to scene, and returns the corresponding object. """
        if self._scn.ngeom >= self._scn.maxgeom:
            raise RuntimeError('Ran out of geoms. maxgeom: %d' % self._scn.maxgeom)

        cdef mjvGeom *g = self._scn.geoms + self._scn.ngeom

        # default values.
        g.dataid = -1
        g.objtype = const.OBJ_UNKNOWN
        g.objid = -1
        g.category = const.CAT_DECOR
        g.texid = -1
        g.texuniform = 0
        g.texrepeat[0] = 1
        g.texrepeat[1] = 1
        g.emission = 0
        g.specular = 0.5
        g.shininess = 0.5
        g.reflectance = 0
        g.type = const.GEOM_BOX
        g.size[:] = np.ones(3) * 0.1
        g.mat[:] = np.eye(3).flatten()
        g.rgba[:] = np.ones(4)
        wrapped = WrapMjvGeom(g)

        for key, value in marker_params.items():
            if isinstance(value, (int, float)):
                setattr(wrapped, key, value)
            elif isinstance(value, np.ndarray):
                getattr(wrapped, key)[:] = value[:]
            elif isinstance(value, str):
                assert key == "label", "Only label is a string in mjvGeom."
                if value == None:
                    g.label[0] = 0
                else:
                    strncpy(g.label, value.encode(), 100)
            else:
                raise ValueError("mjvGeom doesn't have field %s" % key)

        self._scn.ngeom += 1


    def __dealloc__(self):
        mjr_freeContext(&self._con)
        mjv_freeScene(&self._scn)


class MjRenderContextOffscreen(MjRenderContext):

    def __cinit__(self, MjSim sim):
        super().__init__(sim, offscreen=True)

class MjRenderContextWindow(MjRenderContext):

    def __init__(self, MjSim sim):
        super().__init__(sim, offscreen=False)

        assert isinstance(self._opengl_context, GlfwContext), (
            "Only GlfwContext supported for windowed rendering")

    @property
    def window(self):
        return self._opengl_context.window

    def render(self):
        if self.window is None or glfw.window_should_close(self.window):
            return

        glfw.make_context_current(self.window)
        super().render(*glfw.get_framebuffer_size(self.window))
        glfw.swap_buffers(self.window)



include "mjmodel.pxd"


cdef extern from "mjvisualize.h" nogil:

    #---------------------------- global constants -----------------------------------------

    enum: mjNGROUP                    # number of geom and site groups with visflags
    enum: mjMAXOVERLAY                # maximum number of characters in overlay text
    enum: mjMAXLINE                   # maximum number of lines per plot
    enum: mjMAXLINEPNT                # maximum number points per line
    enum: mjMAXPLANEGRID              # maximum number of grid points for plane rendering


    #------------------------------- 3D visualization --------------------------------------

    ctypedef enum mjtCatBit:             # bitflags for mjvGeom category
        mjCAT_STATIC        = 1,        # model elements in body 0
        mjCAT_DYNAMIC       = 2,        # model elements in all other bodies
        mjCAT_DECOR         = 4,        # decorative geoms
        mjCAT_ALL           = 7         # select all categories


    ctypedef enum mjtMouse:              # mouse interaction mode
        mjMOUSE_NONE         = 0,       # no action
        mjMOUSE_ROTATE_V,               # rotate, vertical plane
        mjMOUSE_ROTATE_H,               # rotate, horizontal plane
        mjMOUSE_MOVE_V,                 # move, vertical plane
        mjMOUSE_MOVE_H,                 # move, horizontal plane
        mjMOUSE_ZOOM,                   # zoom
        mjMOUSE_SELECT                  # selection


    ctypedef enum mjtPertBit:            # mouse perturbations
        mjPERT_TRANSLATE    = 1,        # translation
        mjPERT_ROTATE       = 2         # rotation


    ctypedef enum mjtCamera:             # abstract camera type
        mjCAMERA_FREE        = 0,       # free camera
        mjCAMERA_TRACKING,              # tracking camera; uses trackbodyid
        mjCAMERA_FIXED,                 # fixed camera; uses fixedcamid
        mjCAMERA_USER                   # user is responsible for setting OpenGL camera


    ctypedef enum mjtLabel:              # object labeling
        mjLABEL_NONE        = 0,        # nothing
        mjLABEL_BODY,                   # body labels
        mjLABEL_JOINT,                  # joint labels
        mjLABEL_GEOM,                   # geom labels
        mjLABEL_SITE,                   # site labels
        mjLABEL_CAMERA,                 # camera labels
        mjLABEL_LIGHT,                  # light labels
        mjLABEL_TENDON,                 # tendon labels
        mjLABEL_ACTUATOR,               # actuator labels
        mjLABEL_CONSTRAINT,             # constraint labels
        mjLABEL_SELECTION,              # selected object
        mjLABEL_SELPNT,                 # coordinates of selection point
        mjLABEL_CONTACTFORCE,           # magnitude of contact force

    enum: mjNLABEL                        # number of label types


    ctypedef enum mjtFrame:              # frame visualization
        mjFRAME_NONE        = 0,        # no frames
        mjFRAME_BODY,                   # body frames
        mjFRAME_GEOM,                   # geom frames
        mjFRAME_SITE,                   # site frames
        mjFRAME_CAMERA,                 # camera frames
        mjFRAME_LIGHT,                  # light frames
        mjFRAME_WORLD,                  # world frame

    enum: mjNFRAME                        # number of visualization frames


    ctypedef enum mjtVisFlag:            # flags enabling model element visualization
        mjVIS_CONVEXHULL    = 0,        # mesh convex hull
        mjVIS_TEXTURE,                  # textures
        mjVIS_JOINT,                    # joints
        mjVIS_ACTUATOR,                 # actuators
        mjVIS_CAMERA,                   # cameras
        mjVIS_LIGHT,                    # lights
        mjVIS_CONSTRAINT,               # point constraints
        mjVIS_INERTIA,                  # equivalent inertia boxes
        mjVIS_PERTFORCE,                # perturbation force
        mjVIS_PERTOBJ,                  # perturbation object
        mjVIS_CONTACTPOINT,             # contact points
        mjVIS_CONTACTFORCE,             # contact force
        mjVIS_CONTACTSPLIT,             # split contact force into normal and tanget
        mjVIS_TRANSPARENT,              # make dynamic geoms more transparent
        mjVIS_AUTOCONNECT,              # auto connect joints and body coms
        mjVIS_COM,                      # center of mass
        mjVIS_SELECT,                   # selection point
        mjVIS_STATIC,                   # static bodies

    enum: mjNVISFLAG                      # number of visualization flags


    ctypedef enum mjtRndFlag:            # flags enabling rendering effects
        mjRND_SHADOW        = 0,        # shadows
        mjRND_WIREFRAME,                # wireframe
        mjRND_REFLECTION,               # reflections
        mjRND_FOG,                      # fog
        mjRND_SKYBOX,                   # skybox

    enum: mjNRNDFLAG                      # number of rendering flags


    ctypedef enum mjtStereo:             # type of stereo rendering
        mjSTEREO_NONE       = 0,        # no stereo; use left eye only
        mjSTEREO_QUADBUFFERED,          # quad buffered; revert to side-by-side if no hardware support
        mjSTEREO_SIDEBYSIDE             # side-by-side


    ctypedef struct mjvPerturb:                  # object selection and perturbation
        int      select                 # selected body id; non-positive: none
        int      active                 # perturbation bitmask (mjtPertBit)
        mjtNum   refpos[3]              # desired position for selected object
        mjtNum   refquat[4]             # desired orientation for selected object
        mjtNum   localpos[3]            # selection point in object coordinates
        mjtNum   scale                  # relative mouse motion-to-space scaling (set by initPerturb)


    ctypedef struct mjvCamera:                   # abstract camera
        # type and ids
        int      type                   # camera type (mjtCamera)
        int      fixedcamid             # fixed camera id
        int      trackbodyid            # body id to track

        # abstract camera pose specification
        mjtNum   lookat[3]              # lookat point
        mjtNum   distance               # distance to lookat point or tracked body
        mjtNum   azimuth                # camera azimuth (deg)
        mjtNum   elevation              # camera elevation (deg)


    ctypedef struct mjvGLCamera:                 # OpenGL camera
        # camera frame
        float    pos[3]                 # position
        float    forward[3]             # forward direction
        float    up[3]                  # up direction

        # camera projection
        float    frustum_center         # hor. center (left,right set to match aspect)
        float    frustum_bottom         # bottom
        float    frustum_top            # top
        float    frustum_near           # near
        float    frustum_far            # far


    ctypedef struct mjvGeom:                     # abstract geom
        # type info
        int      type                   # geom type (mjtGeom)
        int      dataid                 # mesh, hfield or plane id; -1: none
        int      objtype                # mujoco object type; mjOBJ_UNKNOWN for decor
        int      objid                  # mujoco object id; -1 for decor
        int      category               # visual category
        int      texid                  # texture id; -1: no texture
        int      texuniform             # uniform cube mapping

        # OpenGL info
        float    texrepeat[2]           # texture repetition for 2D mapping
        float    size[3]                # size parameters
        float    pos[3]                 # Cartesian position
        float    mat[9]                 # Cartesian orientation
        float    rgba[4]                # color and transparency
        float    emission               # emission coef
        float    specular               # specular coef
        float    shininess              # shininess coef
        float    reflectance            # reflectance coef
        char     label[100]             # text label

        # transparency rendering (set internally)
        float    camdist                # distance to camera (used by sorter)
        float    modelrbound            # geom rbound from model, 0 if not model geom
        mjtByte  transparent            # treat geom as transparent


    ctypedef struct mjvLight:                    # OpenGL light
        float    pos[3]                 # position rel. to body frame
        float    dir[3]                 # direction rel. to body frame
        float    attenuation[3]         # OpenGL attenuation (quadratic model)
        float    cutoff                 # OpenGL cutoff
        float    exponent               # OpenGL exponent
        float    ambient[3]             # ambient rgb (alpha=1)
        float    diffuse[3]             # diffuse rgb (alpha=1)
        float    specular[3]            # specular rgb (alpha=1)
        mjtByte  headlight              # headlight
        mjtByte  directional            # directional light
        mjtByte  castshadow             # does light cast shadows


    ctypedef struct mjvOption:                   # abstract visualization options
        int      label                  # what objects to label (mjtLabel)
        int      frame                  # which frame to show (mjtFrame)
        mjtByte  geomgroup[mjNGROUP]    # geom visualization by group
        mjtByte  sitegroup[mjNGROUP]    # site visualization by group
        mjtByte  flags[mjNVISFLAG]      # visualization flags (indexed by mjtVisFlag)


    ctypedef struct mjvScene:                    # abstract scene passed to OpenGL renderer
        # abstract geoms
        int      maxgeom                # size of allocated geom buffer
        int      ngeom                  # number of geoms currently in buffer
        mjvGeom* geoms                  # buffer for geoms
        int*     geomorder              # buffer for ordering geoms by distance to camera

        # OpenGL lights
        int      nlight                 # number of lights currently in buffer
        mjvLight lights[8]              # buffer for lights

        # OpenGL cameras
        mjvGLCamera camera[2]           # left and right camera

        # OpenGL model transformation
        mjtByte  enabletransform        # enable model transformation
        float    translate[3]           # model translation
        float    rotate[4]              # model quaternion rotation
        float    scale                  # model scaling

        # OpenGL rendering effects
        int      stereo                 # stereoscopic rendering (mjtStereo)
        mjtByte  flags[mjNRNDFLAG]      # rendering flags (indexed by mjtRndFlag)



    ctypedef struct mjvFigure:          # abstract 2D figure passed to OpenGL renderer
        # enable/disable flags
        int     flg_legend              # show legend
        int     flg_ticklabel[2]        # show grid tick labels (x,y)
        int     flg_extend              # automatically extend axis ranges to fit data
        int     flg_barplot             # isolated line segments (i.e. GL_LINES)

        # figure options
        int     gridsize[2]             # number of grid points in (x,y)
        float   gridrgb[3]              # grid line rgb
        float   gridwidth               # grid line width
        float   figurergba[4]           # figure color and alpha
        float   legendrgba[4]           # legend color and alpha
        float   textrgb[3]              # text color
        float   range[2][2]             # axis ranges; (min>=max) automatic
        char    xlabel[100]             # x-axis label
        char    title[100]              # figure title
        char    xformat[20]             # x-tick label format for sprintf
        char    yformat[20]             # y-tick label format for sprintf
        char    minwidth[20]            # string used to determine min y-tick width

        # line data
        int     linepnt[mjMAXLINE]                   # number of points in line; (0) disable
        float   linergb[mjMAXLINE][3]                # line color
        float   linewidth[mjMAXLINE]                 # line width
        float   linedata[mjMAXLINE][2*mjMAXLINEPNT]  # line data (x,y)
        char    linename[mjMAXLINE][100]             # line name for legend





class MjSimState(namedtuple('SimStateBase', 'time qpos qvel act udd_state')):
    """ Represents the full state of the simulator."""
    __slots__ = ()

    # need to implement this because numpy doesn't support == on arrays
    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented

        if set(self.udd_state.keys()) != set(other.udd_state.keys()):
            return False

        for k in self.udd_state.keys():
            if isinstance(self.udd_state[k], Number) and self.udd_state[k] != other.udd_state[k]:
                return False
            elif not np.array_equal(self.udd_state[k], other.udd_state[k]):
                return False

        return (self.time == other.time and
                np.array_equal(self.qpos, other.qpos) and
                np.array_equal(self.qvel, other.qvel) and
                np.array_equal(self.act, other.act))

    def __ne__(self, other):
        return not self.__eq__(other)

    def flatten(self):
        """ Flattens a state into a numpy array of numbers."""
        if self.act is None:
            act = np.empty(0)
        else:
            act = self.act
        state_tuple = ([self.time], self.qpos, self.qvel, act,
                       MjSimState._flatten_dict(self.udd_state))
        return np.concatenate(state_tuple)

    @staticmethod
    def _flatten_dict(d):
        a = []
        for k in sorted(d.keys()):
            v = d[k]
            if isinstance(v, Number):
                a.extend([v])
            else:
                a.extend(v.ravel())

        return np.array(a)

    @staticmethod
    def from_flattened(array, sim):
        idx_time = 0
        idx_qpos = idx_time + 1
        idx_qvel = idx_qpos + sim.model.nq
        idx_act = idx_qvel + sim.model.nv
        idx_udd = idx_act + sim.model.na

        time = array[idx_time]
        qpos = array[idx_qpos:idx_qpos + sim.model.nq]
        qvel = array[idx_qvel:idx_qvel + sim.model.nv]
        if sim.model.na == 0:
            act = None
        else:
            act = array[idx_act:idx_act + sim.model.na]
        flat_udd_state = array[idx_udd:]
        udd_state = MjSimState._unflatten_dict(flat_udd_state, sim.udd_state)

        return MjSimState(time, qpos, qvel, act, udd_state)

    @staticmethod
    def _unflatten_dict(a, schema_example):
        d = {}
        idx = 0
        for k in sorted(schema_example.keys()):
            schema_val = schema_example[k]
            if isinstance(schema_val, Number):
                val = a[idx]
                idx += 1
                d[k] = val
            else:
                assert isinstance(schema_val, np.ndarray)
                val_array = a[idx:idx+schema_val.size]
                idx += schema_val.size
                val = np.array(val_array).reshape(schema_val.shape)
                d[k] = val
        return d



import os
import sys
from abc import ABCMeta, abstractmethod
from pymj.utils import discover_mujoco


def _add_mujoco_bin_to_dyld_library_path():
    mujoco_path, _ = discover_mujoco()
    bin_path = os.path.join(mujoco_path, "bin")
    old_dyld_library_path = os.getenv("DYLD_LIBRARY_PATH", "")
    os.environ["DYLD_LIBRARY_PATH"] = "{}:{}".format(
        bin_path, old_dyld_library_path)


try:
    _add_mujoco_bin_to_dyld_library_path()
    import glfw
except ImportError:
    pass


class OpenGLContext(metaclass=ABCMeta):

    @abstractmethod
    def make_context_current(self):
        raise NotImplementedError()

    @abstractmethod
    def set_buffer_size(self, width, height):
        raise NotImplementedError()


class GlfwError(RuntimeError):
    pass


class GlfwContext(OpenGLContext):

    _INIT_WIDTH = 1000
    _INIT_HEIGHT = 1000
    _GLFW_IS_INITIALIZED = False

    def __init__(self, offscreen=False):
        GlfwContext._init_glfw()

        self._width = self._INIT_WIDTH
        self._height = self._INIT_HEIGHT
        self.window = self._create_window(offscreen)
        self._set_window_size(self._width, self._height)

    @staticmethod
    def _init_glfw():
        if GlfwContext._GLFW_IS_INITIALIZED:
            return

        if 'glfw' not in globals():
            raise GlfwError("GLFW not installed")

        glfw.set_error_callback(GlfwContext._glfw_error_callback)

        # HAX: sometimes first init() fails, while second works fine.
        glfw.init()
        if not glfw.init():
            raise GlfwError("Failed to initialize GLFW")

        GlfwContext._GLFW_IS_INITIALIZED = True

    def make_context_current(self):
        glfw.make_context_current(self.window)

    def set_buffer_size(self, width, height):
        self._set_window_size(width, height)
        self._width = width
        self._height = height

    def _create_window(self, offscreen):
        if offscreen:
            print("Creating offscreen glfw")
            glfw.window_hint(glfw.VISIBLE, 0)
            glfw.window_hint(glfw.DOUBLEBUFFER, 0)
            init_width, init_height = self._INIT_WIDTH, self._INIT_HEIGHT
        else:
            print("Creating window glfw")
            glfw.window_hint(glfw.SAMPLES, 4)
            glfw.window_hint(glfw.VISIBLE, 1)
            glfw.window_hint(glfw.DOUBLEBUFFER, 1)
            resolution, _, refresh_rate = glfw.get_video_mode(
                glfw.get_primary_monitor())
            init_width, init_height = resolution

        self._width = init_width
        self._height = init_height
        window = glfw.create_window(
            self._width, self._height, "PyMj", None, None)

        if not window:
            raise GlfwError("Failed to create GLFW window")

        return window

    def get_buffer_size(self):
        return glfw.get_framebuffer_size(self.window)

    def _set_window_size(self, target_width, target_height):
        self.make_context_current()
        if target_width != self._width or target_height != self._height:
            self._width = target_width
            self._height = target_height
            glfw.set_window_size(self.window, target_width, target_height)

            # HAX: When running on a Mac with retina screen, the size
            # sometimes doubles
            width, height = glfw.get_framebuffer_size(self.window)
            if target_width != width:
                glfw.set_window_size(self.window, target_width // 2, target_height // 2)

    @staticmethod
    def _glfw_error_callback(error_code, description):
        print("GLFW error (code %d): %s", error_code, description)


class OffscreenOpenGLContext():

    def __init__(self):
        # TODO: support GPU rendering on mac
        # TODO: allow user to pick GPU expilictly from code
        device_id = int(os.getenv('CUDA_VISIBLE_DEVICES', '0').split(',')[0])
        res = initOpenGL(device_id)
        if res != 1:
            raise RuntimeError("Failed to initialize OpenGL")

    def close(self):
        # TODO: properly close OpenGL in our contexts
        closeOpenGL()

    def make_context_current(self):
        # TODO: maybe expose this explicitly?
        pass

    def set_buffer_size(self, int width, int height):
        res = setOpenGLBufferSize(width, height)
        if res != 1:
            raise RuntimeError("Failed to set buffer size")



from tempfile import NamedTemporaryFile
from pymj import functions
from pymj.cymj import PyMjVFS


def test_vfs():
    ''' Test basic VFS functionality '''
    vfs = PyMjVFS()
    functions.mj_defaultVFS(vfs)
    functions.mj_deleteVFS(vfs)


def test_files():
    ''' Testing handling VFS files '''
    vfs = PyMjVFS()
    functions.mj_defaultVFS(vfs)
    with NamedTemporaryFile() as f:
        name = f.name
        # Try to find file before added -> missing
        assert functions.mj_findFileVFS(vfs, name) == -1
        # Add file -> success
        assert functions.mj_addFileVFS(vfs, '', name) == 0
        # Add file again -> failure, duplicate
        assert functions.mj_addFileVFS(vfs, '', name) == 2
        # Find file -> success (index 0)
        assert functions.mj_findFileVFS(vfs, name) == 0
        # Delete file -> success
        assert functions.mj_deleteFileVFS(vfs, name) == 0
        # Delete file again -> failure, missing
        assert functions.mj_deleteFileVFS(vfs, name) == -1
    # Add a file which does not exist -> failure, missing
    assert functions.mj_addFileVFS(vfs, '', name) == -1



import subprocess
import glob
import os.path
import pymj
import sys
pymj_root = os.path.dirname(os.path.dirname(pymj.__file__))


def test_examples():
    scripts = glob.glob("%s/examples/*.py" % pymj_root)
    env = os.environ.update({'TESTING': 'true'})
    assert len(scripts) > 0, 'No example scripts found!'
    for tutorial_script in scripts:
        print("Executing %s" % tutorial_script)

        subprocess.check_call([sys.executable, tutorial_script], env=env)




cdef class MjSimPool(object):
    """
    Keeps a pool of multiple MjSims and enables stepping them quickly
    in parallel.
    """
    # Arrays of pointers to mjDatas and mjModels for fast multithreaded access
    cdef mjModel **_models
    cdef mjData **_datas

    # List of MjSim objects that can be accessed directly.
    cdef readonly list sims
    # Number of .step substeps.
    cdef readonly int nsubsteps

    def __cinit__(self, list sims, int nsubsteps=1):
        """
        Args:
            - sims ([PyMjModel] list): list of sims to add to the pool
            - nsubsteps (int): number of substeps to run on `.step`. The
                individual model `nsubstep`s will be ignored.
        """
        self.sims = sims
        self.nsubsteps = nsubsteps
        self._allocate_data_pointers()

    def reset(self, nsims=None):
        """
        Resets all simulations in pool.
        If nsims is specified, than only the first nsims simulator are reset.
        """
        length = self.nsims

        if nsims is not None:
            if nsims > self.nsims:
                raise ValueError("nsims is larger than pool size")
            length = nsims

        for i in range(length):
            self.sims[i].reset()

    def forward(self, nsims=None):
        """
        Calls `mj_forward` on all simulations in parallel.
        If nsims is specified, than only the first nsims simulator are forwarded.
        """
        cdef int i
        cdef int length = self.nsims

        if nsims is not None:
            if nsims > self.nsims:
                raise ValueError("nsims is larger than pool size")
            length = nsims

        # See explanation in MjSimPool.step() for why we wrap warnings this way
        with wrap_mujoco_warning():
            with nogil, parallel():
                for i in prange(length, schedule='guided'):
                    mj_forward(self._models[i], self._datas[i]);

    def step(self, nsims=None):
        """
        Calls `mj_step` on all simulations in parallel, with `nsubsteps` as
        specified on `MjSimPool` instance creation.
        If nsims is specified, than only the first nsims simulator are stepped.
        """
        cdef int i, j
        cdef int length = self.nsims

        if nsims is not None:
            if nsims > self.nsims:
                raise ValueError("nsims is larger than pool size")
            length = nsims

        for sim in self.sims[:length]:
            sim.step_udd()

        # Wrapping these calls to mj_step is tricky, since they're parallelized
        # and can't access the GIL or global python objects.
        # Because we expect to have fatal warnings, we'll just wrap the entire
        # section, and if any call ends up setting an exception we'll raise.
        with wrap_mujoco_warning():
            with nogil, parallel():
                for i in prange(length, schedule='guided'):
                    for j in range(self.nsubsteps):
                        mj_step(self._models[i], self._datas[i]);

    @property
    def nsims(self):
        """
        Number of simulations in the pool.
        """
        return len(self.sims)

    @staticmethod
    def create_from_sim(sim, nsims):
        sims = [MjSim(sim.model, udd_callback=sim.udd_callback)
                for _ in range(nsims)]
        return MjSimPool(sims, nsubsteps=sim.nsubsteps)

    cdef _allocate_data_pointers(self):
        self._models = <mjModel**>malloc(self.nsims * sizeof(mjModel*))
        self._datas = <mjData**>malloc(self.nsims * sizeof(mjData*))
        for i in range(self.nsims):
            sim = <MjSim> self.sims[i]
            self._models[i] = sim.model.ptr
            self._datas[i] = sim.data.ptr

    def __dealloc__(self):
        free(self._datas)
        free(self._models)


#!/usr/bin/env python3
from setuptools import find_packages, setup
from os.path import join, dirname, realpath
from setuptools.command.install import install

with open(join("pymj", "version.py")) as version_file:
    exec(version_file.read())


class PostInstallCommand(install):
    """Post-installation for installation mode."""

    def run(self):
        install.run(self)
        import pymj  # noqa


def read_requirements_file(filename):
    req_file_path = '%s/%s' % (dirname(realpath(__file__)), filename)
    with open(req_file_path) as f:
        return [line.strip() for line in f]

packages = find_packages()
# Ensure that we don't pollute the global namespace.
for p in packages:
    assert p == 'pymj' or p.startswith('pymj.')

setup(
    name='pymj',
    version=__version__,  # noqa
    author='OpenAI Robotics Team',
    author_email='robotics@openai.com',
    packages=packages,
    include_package_data=True,
    install_requires=read_requirements_file('requirements.txt'),
    tests_require=read_requirements_file('requirements.dev.txt'),
    cmdclass={
        'install': PostInstallCommand,
    },
)



import distutils
import imp
import os
import shutil
import subprocess
import sys
from distutils.core import Extension
from distutils.dist import Distribution
from distutils.sysconfig import customize_compiler
from os.path import abspath, dirname, exists, join, getmtime

import numpy as np
import portalocker
from Cython.Build import cythonize
from Cython.Distutils.old_build_ext import old_build_ext as build_ext

from pymj.utils import discover_mujoco


def load_cython_ext(mjpro_path):
    """
    Loads the cymj Cython extension. This is safe to be called from
    multiple processes running on the same machine.

    Cython only gives us back the raw path, regardless of whether
    it found a cached version or actually compiled. Since we do
    non-idempotent postprocessing of the DLL, be extra careful
    to only do that once and then atomically move to the final
    location.
    """
    if ('glfw' in sys.modules and
            'mujoco' in abspath(sys.modules["glfw"].__file__)):
        print('''
WARNING: Existing glfw python module detected!

MuJoCo comes with its own version of GLFW, so it's preferable to use that one.

The easy solution is to `import pymj` _before_ `import glfw`.
''')

    lockpath = join(dirname(__file__), 'cythonlock_%s.pyc' % (sys.platform))
    lockfile = open(lockpath, 'a')
    max_attempts = 15
    locked = False
    for idx in range(max_attempts):
        try:
            portalocker.lock(lockfile, portalocker.LOCK_EX |
                             portalocker.LOCK_NB)
            locked = True
            break
        except (portalocker.exceptions.LockException, BlockingIOError):
            print("My pid = %s. Failed to acquire lock. Waiting." %
                  os.getpid())
            import time
            time.sleep(10)
    if not locked:
        raise RuntimeError('Failed to acquire pymj.builder lock.')

    try:
        if sys.platform == 'darwin':
            Builder = MacExtensionBuilder
        elif sys.platform == 'linux':
            if exists('/usr/local/nvidia/lib64'):
                Builder = LinuxGPUExtensionBuilder
            else:
                Builder = LinuxCPUExtensionBuilder
        elif sys.platform.startswith("win"):
            Builder = WindowsExtensionBuilder
        else:
            raise RuntimeError("Unsupported platform %s" % sys.platform)

        builder = Builder(mjpro_path)
        os.environ["PATH"] += ":" + join(mjpro_path, "bin")

        cext_so_path = builder.build()

        mod = imp.load_dynamic("cymj", cext_so_path)
        return mod
    finally:
        portalocker.unlock(lockfile)


class custom_build_ext(build_ext):
    """
    Custom build_ext to suppress the "-Wstrict-prototypes" warning.
    It arises from the fact that we're using C++. This seems to be
    the cleanest way to get rid of the extra flag.

    See http://stackoverflow.com/a/36293331/248400
    """

    def build_extensions(self):
        customize_compiler(self.compiler)

        try:
            self.compiler.compiler_so.remove("-Wstrict-prototypes")
        except (AttributeError, ValueError):
            pass
        build_ext.build_extensions(self)


def fix_shared_library(so_file, name, library_path):
    ldd_output = subprocess.check_output(
        ['ldd', so_file]).decode('utf-8')

    if name in ldd_output:
        subprocess.check_call(['patchelf',
                               '--remove-needed',
                               name,
                               so_file])
    subprocess.check_call(
        ['patchelf', '--add-needed',
         library_path,
         so_file])


class MujocoExtensionBuilder():

    CYMJ_DIR_PATH = abspath(dirname(__file__))

    def __init__(self, mjpro_path):
        self.mjpro_path = mjpro_path
        self.extension = Extension(
            'pymj.cymj',
            sources=[join(self.CYMJ_DIR_PATH, "cymj.pyx")],
            include_dirs=[
                self.CYMJ_DIR_PATH,
                join(mjpro_path, 'include'),
                np.get_include(),
            ],
            libraries=['mujoco150'],
            library_dirs=[join(mjpro_path, 'bin')],
            extra_compile_args=[
                '-fopenmp',  # needed for OpenMP
                '-w',  # suppress numpy compilation warnings
            ],
            extra_link_args=['-fopenmp'],
            language='c')

    def build(self):
        dist = Distribution({
            "script_name": None,
            "script_args": ["build_ext"]
        })
        dist.ext_modules = cythonize([self.extension])
        dist.include_dirs = []
        dist.cmdclass = {'build_ext': custom_build_ext}
        build = dist.get_command_obj('build')
        # following the convention of cython's pyxbuild and naming
        # base directory "_pyxbld"
        build.build_base = join(self.CYMJ_DIR_PATH, 'generated',
                                '_pyxbld_%s' % self.__class__.__name__)
        dist.parse_command_line()
        obj_build_ext = dist.get_command_obj("build_ext")
        dist.run_commands()
        so_file_path, = obj_build_ext.get_outputs()
        return so_file_path


class WindowsExtensionBuilder(MujocoExtensionBuilder):

    def __init__(self, mjpro_path):
        super().__init__(mjpro_path)
        self.extension.sources.append(self.CYMJ_DIR_PATH + "/gl/dummyshim.c")


class LinuxCPUExtensionBuilder(MujocoExtensionBuilder):

    def __init__(self, mjpro_path):
        super().__init__(mjpro_path)

        self.extension.sources.append(
            join(self.CYMJ_DIR_PATH, "gl", "osmesashim.c"))
        self.extension.libraries.extend(['glewosmesa', 'OSMesa'])
        self.extension.runtime_library_dirs = [join(mjpro_path, 'bin')]


class LinuxGPUExtensionBuilder(MujocoExtensionBuilder):

    def __init__(self, mjpro_path):
        super().__init__(mjpro_path)

        self.extension.sources.append(self.CYMJ_DIR_PATH + "/gl/eglshim.c")
        self.extension.include_dirs.append(self.CYMJ_DIR_PATH + '/vendor/egl')
        self.extension.libraries.extend(['glewegl'])
        self.extension.runtime_library_dirs = [join(mjpro_path, 'bin')]

    def build(self):
        so_file_path = super().build()
        nvidia_lib_dir = '/usr/local/nvidia/lib64/'
        fix_shared_library(so_file_path, 'libOpenGL.so',
                           join(nvidia_lib_dir, 'libOpenGL.so.0'))
        fix_shared_library(so_file_path, 'libEGL.so',
                           join(nvidia_lib_dir, 'libEGL.so.1'))
        return so_file_path


class MacExtensionBuilder(MujocoExtensionBuilder):

    def __init__(self, mjpro_path):
        super().__init__(mjpro_path)

        self.extension.sources.append(self.CYMJ_DIR_PATH + "/gl/dummyshim.c")
        self.extension.libraries.extend(['glfw.3'])
        self.extension.define_macros = [('ONMAC', None)]
        self.extension.runtime_library_dirs = [join(mjpro_path, 'bin')]

    def build(self):
        # Prefer GCC 6 for now since GCC 7 may behave differently.
        c_compilers = ['/usr/local/bin/gcc-6', '/usr/local/bin/gcc-7']
        available_c_compiler = None
        for c_compiler in c_compilers:
            if distutils.spawn.find_executable(c_compiler) is not None:
                available_c_compiler = c_compiler
                break
        if available_c_compiler is None:
            raise RuntimeError(
                'Could not find GCC 6 or GCC 7 executable.\n\n'
                'HINT: On OS X, install GCC 6 with '
                '`brew install gcc --without-multilib`.')
        os.environ['CC'] = available_c_compiler

        so_file_path = super().build()
        del os.environ['CC']
        return self.manually_link_libraries(so_file_path)

    def manually_link_libraries(self, raw_cext_dll_path):
        root, ext = os.path.splitext(raw_cext_dll_path)
        final_cext_dll_path = root + '_final' + ext

        # If someone else already built the final DLL, don't bother
        # recreating it here, even though this should still be idempotent.
        if (exists(final_cext_dll_path) and
                getmtime(final_cext_dll_path) >= getmtime(raw_cext_dll_path)):
            return final_cext_dll_path

        tmp_final_cext_dll_path = final_cext_dll_path + '~'
        shutil.copyfile(raw_cext_dll_path, tmp_final_cext_dll_path)

        mj_bin_path = join(self.mjpro_path, 'bin')

        # Fix the rpath of the generated library -- i lost the Stackoverflow
        # reference here
        from_mujoco_path = '@executable_path/libmujoco150.dylib'
        to_mujoco_path = '%s/libmujoco150.dylib' % mj_bin_path
        subprocess.check_call(['install_name_tool',
                               '-change',
                               from_mujoco_path,
                               to_mujoco_path,
                               tmp_final_cext_dll_path])

        from_glfw_path = 'libglfw.3.dylib'
        to_glfw_path = os.path.join(mj_bin_path, 'libglfw.3.dylib')
        subprocess.check_call(['install_name_tool',
                               '-change',
                               from_glfw_path,
                               to_glfw_path,
                               tmp_final_cext_dll_path])

        os.rename(tmp_final_cext_dll_path, final_cext_dll_path)
        return final_cext_dll_path


class MujocoException(Exception):
    pass


def user_warning_raise_exception(warn_bytes):
    '''
    User-defined warning callback, which is called by mujoco on warnings.
    Here we have two primary jobs:
        - Detect known warnings and suggest fixes (with code)
        - Decide whether to raise an Exception and raise if needed
    More cases should be added as we find new failures.
    '''
    # TODO: look through test output to see MuJoCo warnings to catch
    # and recommend. Also fix those tests
    warn = warn_bytes.decode()  # Convert bytes to string
    if 'Pre-allocated constraint buffer is full' in warn:
        raise MujocoException(warn + 'Increase njmax in mujoco XML')
    if 'Pre-allocated contact buffer is full' in warn:
        raise MujocoException(warn + 'Increase njconmax in mujoco XML')
    raise MujocoException('Got MuJoCo Warning: {}'.format(warn))


def user_warning_ignore_exception(warn_bytes):
    pass


class ignore_mujoco_warnings:
    """
    Class to turn off mujoco warning exceptions within a scope. Useful for
    large, vectorized rollouts.
    """

    def __enter__(self):
        self.prev_user_warning = cymj.get_warning_callback()
        cymj.set_warning_callback(user_warning_ignore_exception)
        return self

    def __exit__(self, type, value, traceback):
        cymj.set_warning_callback(self.prev_user_warning)


mjpro_path, key_path = discover_mujoco()
cymj = load_cython_ext(mjpro_path)


# Trick to expose all mj* functions from mujoco in pymj.*
class dict2(object):
    pass


functions = dict2()
for func_name in dir(cymj):
    if func_name.startswith("_mj"):
        setattr(functions, func_name[1:], getattr(cymj, func_name))

functions.mj_activate(key_path)

# Set user-defined callbacks that raise assertion with message
cymj.set_warning_callback(user_warning_raise_exception)
